

from typing import Any, Type
import logging

from ._coder import Coder
from ._key import KeyGen
from ._detaBase import DetaBase,SyncBase,AsyncBase
from ._helpers import getCurrentTimestamp,checkExpiredTimestamp


logger = logging.getLogger(__name__)

class Cache:
    def __init__(self,
        db: DetaBase,
        expire: int,
        function: Any,
        args: tuple,
        kwargs: dict,
        keyGen: Type[KeyGen],
        coder: Type[Coder],
    ):
        self.db = db
        self.expire = expire
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.key = keyGen.get(self.function, self.args, self.kwargs)
        self.coder = coder
    
    def getCached(self):
        '''create `cached` variable'''
        raise NotImplementedError
    
    def putCache(self):
        '''put `cached`'''
        raise NotImplementedError
    
    def functionCall(self):
        '''create `fucResponse` variable'''
        raise NotImplementedError
    
    def checkCached(self):
        raise NotImplementedError
    
    def putValue(self,data: Any) -> dict:
        return {
                'value': self.coder.encode(data),
                'type': type(data).__name__,
                'expire': self.expire,
                'timestamp': getCurrentTimestamp()
            }
    
    def checkIfCachedOrExpired(self,cached:dict):
        if not cached:
            logger.info(f'{self.function.__name__} function has no cache..')
            return True
        if cached.get('expire') != self.expire:
                logger.info(f'{self.function.__name__} function updating.... expire time ')
                return True
        if self.expire and checkExpiredTimestamp(
                cached.get('expire'), cached.get('timestamp'), getCurrentTimestamp()):
                logger.info(f'{self.function.__name__} function cache expire..')
                return True

class SyncCache(Cache):
    def __init__(self, db: SyncBase, expire: int, function: Any, args: tuple, kwargs: dict, keyGen: Type[KeyGen], coder: Type[Coder]):
        super().__init__(db, expire, function, args, kwargs, keyGen, coder)
    
    def getCache(self):
        self.cached = self.db.get(key=self.key)
        return self.cached
    
    def putCache(self,data: Any):
        self.db.put(data=self.putValue(data), key=self.key)
        logger.info(f'{self.function.__name__} function cached..')
        return data
        
    def functionCall(self):
        return self.function(*self.args, **self.kwargs)
    
    def checkCached(self):
        if self.checkIfCachedOrExpired(self.getCache()):
            return self.putCache(self.functionCall())
        logger.info(f'{self.function.__name__} function cached HIT')
        return self.coder.decode(self.cached)


class AsyncCache(Cache):
    def __init__(self, db: AsyncBase, expire: int, function: Any, args: tuple, kwargs: dict, keyGen: Type[KeyGen], coder: Type[Coder]):
        super().__init__(db, expire, function, args, kwargs, keyGen, coder)
    
    async def getCache(self):
        self.cached = await self.db.get(key=self.key)
        return self.cached
    
    async def putCache(self,data: Any):
        await self.db.put(data=self.putValue(data), key=self.key)
        logger.info(f'{self.function.__name__} function cached..')
        return data
    
    async def functionCall(self):
        return await self.function(*self.args, **self.kwargs)
    
    async def checkCached(self):
        if self.checkIfCachedOrExpired(await self.getCache()):
            return await self.putCache(await self.functionCall())
        logger.info(f'{self.function.__name__} function cached HIT')
        return self.coder.decode(self.cached)






import typing
import aiohttp
import requests


class DetaBase:
    def __init__(self,projectKey: str ,baseName: str , projectId: str = None):
        if not "_" in projectKey:
            raise ValueError("Bad project key provided")

        if not projectId:
            projectId = projectKey.split("_")[0]
        
        self._headers = {
                "Content-type": "application/json",
                "X-API-Key": projectKey,
            }
        self._baseUrl = f'https://database.deta.sh/v1/{projectId}/{baseName}'

    def put(self, data: dict) -> dict:
        raise NotImplementedError

    def get(self, key: str) -> dict:
        raise NotImplementedError

class SyncBase(DetaBase):
    def __init__(self, projectKey: str, baseName: str, projectId: str = None):
        super().__init__(projectKey, baseName, projectId)

    def put(self, data: typing.Union[dict, list, str, int, bool],key: str = None) -> dict:
        if key:
            data["key"] = key
        with requests.Session() as _session:
            with _session.put(f"{self._baseUrl}/items", json={"items": [data]},headers=self._headers) as resp:
                return resp.json()["processed"]["items"][0]

    def get(self, key: str) -> dict:
        with requests.Session() as _session:
            with _session.get(f"{self._baseUrl}/items/{key}",headers=self._headers) as resp:
                _res = resp.json()
                return _res if len(_res) > 1 else None

class AsyncBase(DetaBase):
    def __init__(self, projectKey: str, baseName: str, projectId: str = None):
        super().__init__(projectKey, baseName, projectId)

    async def put(self, data: typing.Union[dict, list, str, int, bool],key: str = None)-> dict:
        if key:
            data["key"] = key
        async with aiohttp.ClientSession(headers=self._headers) as _session:
            async with _session.put(f"{self._baseUrl}/items", json={"items": [data]}) as resp:
                _res =  await resp.json()
                return _res["processed"]["items"][0]

    async def get(self, key: str)-> dict:
        async with aiohttp.ClientSession(headers=self._headers) as _session:
            async with _session.get(f"{self._baseUrl}/items/{key}") as resp:
                _res = await resp.json()
                return _res if len(_res) > 1 else None





import json
import pickle
from typing import Any
from fastapi.encoders import jsonable_encoder

from ._constants import JSON_CONVERTERS,FASTAPI_CONVERTERS


class Coder:
    @classmethod
    def encode(cls, value: Any):
        raise NotImplementedError

    @classmethod
    def decode(cls, value: Any):
        raise NotImplementedError


class JsonCoder(Coder):
    @classmethod
    def encode(cls, value: Any):
        return json.dumps(value)

    @classmethod
    def decode(cls, value: Any):
        return json.loads(value)


class PickleCoder(Coder):
    @classmethod
    def encode(cls, value: Any):
        return pickle.dumps(value)

    @classmethod
    def decode(cls, value: Any):
        return pickle.loads(value)


def objDecode(value,con):
    _type = value.get("type")
    if not _type:
        return value

    if _type in con:
        return con[_type](value["value"])
    else:
        raise TypeError("Unknown {}".format(_type))

class DetaCoder(Coder):
    '''(dict, list, tuple, set, float, str, int, bool)'''
    @classmethod
    def encode(cls, value: Any):
        if isinstance(value, (dict, list, float, str, int, bool)):
            return value
        elif isinstance(value, (tuple, set)):
            return list(value)
        else:
            raise Exception(
                "function response must be (dict, list, tuple, set, float, str, int, bool)")

    @classmethod
    def decode(cls, value: Any):
        return objDecode(value,JSON_CONVERTERS)


class FastAPICoder(Coder):
    @classmethod
    def encode(cls, value: Any):
        return jsonable_encoder(value)

    @classmethod
    def decode(cls, value: Any):
        return objDecode(value,FASTAPI_CONVERTERS)






import inspect
from fastapi.requests import Request

from ._helpers import createStringHashKey


class KeyGen:
    @classmethod
    def get(cls, function, args: tuple, kwargs: dict) -> str:
        raise NotImplementedError


class JsonKeyGen(KeyGen):
    @classmethod
    def get(cls,function, args: tuple, kwargs: dict):
        '''Returns a deta cache key'''

        argspec = inspect.getfullargspec(function)

        data = {str(argspec.args[index]): str(arg) if isinstance(arg,
                (dict, list, tuple, set, str, int, bool)) else None for index, arg in enumerate(args)
                if argspec.args and type(argspec.args is list)}

        data.update({str(k): str(v) if isinstance(v, (dict, list, tuple, set, str, int, bool))
                    else None for k, v in kwargs.items() if kwargs})
        
        return createStringHashKey(f'{function.__name__}{data}')

class DetaKeyGen(JsonKeyGen):
    pass

class FastAPIKeyGen(KeyGen):
    @classmethod
    def get(cls,function, args: tuple, kwargs: dict,):
        '''Returns a deta cache key'''

        request = kwargs.get('request')

        assert request, f"function {function.__name__} needs a `request` argument"

        if not isinstance(request, Request):
            raise Exception(
                "`request` must be an instance of `starlette.request.Request`")

        return createStringHashKey(f'{function.__name__}{request.url}')




import asyncio
from functools import wraps
from typing import Type

from ._coder import Coder,JsonCoder,FastAPICoder,DetaCoder
from ._key import KeyGen,JsonKeyGen,FastAPIKeyGen,DetaKeyGen
from ._detaBase import SyncBase,AsyncBase
from ._cache import AsyncCache,SyncCache



class BaseDecorator:

    def __init__(self, 
                projectKey: str = None, 
                projectId: str = None, 
                baseName: str = 'cache',
                keyGen:Type[KeyGen] = None,
                coder:Type[Coder] = None,
                ):
        self._syncDb = SyncBase(projectKey, baseName,projectId)
        self._asyncDb = AsyncBase(projectKey, baseName,projectId)
        self.keyGen = keyGen
        self.coder=coder

    def cache(self, 
            expire: int = 0,
            keyGen:Type[KeyGen]=None,
            coder:Type[Coder]=None,
            ) -> None:

        def wrapped(function):
            @wraps(function)
            async def asyncWrappedFunction(*args, **kwargs):
                return await AsyncCache(
                    self._asyncDb,
                    expire,
                    function,
                    args,
                    kwargs,
                    keyGen if keyGen else self.keyGen,
                    coder if coder else self.coder,
                ).checkCached()

            @wraps(function)
            def syncWrappedFunction(*args, **kwargs):
                return SyncCache(
                    self._syncDb,
                    expire,
                    function,
                    args,
                    kwargs,
                    keyGen if keyGen else self.keyGen,
                    coder if coder else self.coder,
                ).checkCached()

            if asyncio.iscoroutinefunction(function):
                return asyncWrappedFunction
            else:
                return syncWrappedFunction
        return wrapped



class DetaCache(BaseDecorator):
    def __init__(self, 
                projectKey: str = None,
                projectId: str = None, 
                baseName: str = 'cache',
                keyGen:Type[KeyGen]=DetaKeyGen,
                coder:Type[Coder]=DetaCoder,
            ):
        super().__init__(projectKey, projectId, baseName,keyGen,coder)

class JsonCache(BaseDecorator):
    def __init__(self, 
                projectKey: str = None,
                projectId: str = None, 
                baseName: str = 'cache',
                keyGen:Type[KeyGen]=JsonKeyGen,
                coder:Type[Coder]=JsonCoder,
            ):
        super().__init__(projectKey, projectId, baseName,keyGen,coder)


class FastAPICache(BaseDecorator):
    def __init__(self, 
                projectKey: str = None,
                projectId: str = None, 
                baseName: str = 'cache',
                keyGen:Type[KeyGen]=FastAPIKeyGen,
                coder:Type[Coder]=FastAPICoder,
            ):
        super().__init__(projectKey, projectId, baseName,keyGen,coder)




from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name='detacache',
    packages=['detacache'],
    version='v0.1.1',
    license='MIT',
    description='Async and Sync Function Decorator to cache function call\'s to Deta base',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='vidya sagar',
    author_email='svidya051@gmail.com',
    url='https://github.com/vidyasagar1432/detacache',
    keywords=['deta', 'cache', 'asyncio', 'deta base cache','fastapi cache',
              'cache api call', 'cache functions', 'cache requests'],
    install_requires=[
        'aiohttp==3.8.1',
        'requests==2.26.0',
        'fastapi==0.74.0'
    ],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
    ],
    python_requires='>=3.6',
)




import hashlib
import datetime
import logging
import sys


logger = logging.getLogger('detacache')
logger.setLevel(logging.INFO)
handler = logging.StreamHandler(sys.stderr)
handler.setLevel(logging.INFO)
formatter = logging.Formatter(
    '%(levelname)s:%(module)s:%(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)


def createIntHashKey(string: str):
    '''Returns a md5 Hash of string as `int`'''
    return int(hashlib.md5(str(string).encode()).hexdigest(), 16)


def createStringHashKey(string: str):
    '''Returns a md5 Hash of string as `string`'''
    return hashlib.md5(str(string).encode()).hexdigest()


def getCurrentTimestamp():
    '''Returns Current Timestamp as `int`'''
    return int(round(datetime.datetime.now().timestamp()))


def checkExpiredTimestamp(expire: int, initialTimestamp: int, currentTimestamp: int):
    '''Returns `True` if expired else `False`'''
    return initialTimestamp + expire < currentTimestamp





from fastapi.responses import HTMLResponse, PlainTextResponse, Response

JSON_CONVERTERS = {
    "dict": lambda x: dict(x),
    "list": lambda x: list(x),
    "tuple": lambda x: tuple(x),
    "float": lambda x: float(x),
    "set": lambda x: set(x),
    "str": lambda x: str(x),
    "int": lambda x: int(x),
    "bool": lambda x: bool(x),
}


_FASTAPIHTML= lambda x: HTMLResponse(str(x['body']), headers=dict(x['raw_headers']), status_code=x['status_code'])


FASTAPI_CONVERTERS = {
    "_TemplateResponse": _FASTAPIHTML,
    "HTMLResponse": _FASTAPIHTML,
    "PlainTextResponse": lambda x: PlainTextResponse(str(x['body']), headers=dict(x['raw_headers']), status_code=x['status_code']),
    "Response": lambda x: Response(str(x['body']), headers=dict(x['raw_headers']), status_code=x['status_code']),
}

FASTAPI_CONVERTERS.update(JSON_CONVERTERS)
